#include "string.h"
#include "printf.h"
#include "failure.h"
#include "page.h"
#include "multiboot.h"
#include "system.h"
#include "msr.h"
#include "elf.h"
#include "vm.h"
#include "svm.h"
#include "alloc.h"
#include "vmexit.h"


#include "vmm.h" /* [DEBUG] */


static struct vmcb *
alloc_vmcb ( void )
{
	struct vmcb *vmcb;

	const unsigned long pfn = alloc_pages ( 1, 1 );
	vmcb = ( struct vmcb * ) VIRT ( pfn << PAGE_SHIFT );
	memset ( ( char * ) vmcb, 0, sizeof ( struct vmcb ) );

	return vmcb;
}

static unsigned long 
create_intercept_table ( unsigned long size )
{
	const unsigned long pfn = alloc_pages ( size >> PAGE_SHIFT, 1 );
	void *p = ( void * ) VIRT ( pfn << PAGE_SHIFT );

	/* vol. 2, p. 445 */
	memset ( p, 0xff, size );

	return pfn << PAGE_SHIFT; 
}

static unsigned long
alloc_vm_pmem ( unsigned long size )
{
	const unsigned long align = 1 << ( PAGE_SHIFT_2MB - PAGE_SHIFT );
	const unsigned long pfn = alloc_pages ( size >> PAGE_SHIFT, align );
	return ( unsigned long ) VIRT ( pfn << PAGE_SHIFT );
}

static struct multiboot_info *
init_vm_mbi ( unsigned long vm_pmem_start )
{
	enum { INSTALL_PADDR = 0x2d0e0UL }; /* < 1 MB */
	struct multiboot_info *mbi = ( struct multiboot_info *) ( vm_pmem_start + INSTALL_PADDR );

	/* [TODO] */

	printf ( "Multiboot information initialized.\n" );

	return mbi;
}

static int
is_reserved_pmem ( unsigned long pfn )
{
	return ( pfn < ( 1 << ( 20 - PAGE_SHIFT ) ) );
}

/* ۥޥʪ¥ޥʪбŤڡơ֥
 * ʪɥ쥹֤*/
static unsigned long 
create_vm_pmem_mapping_table ( unsigned long vm_pmem_start, unsigned long vm_pmem_size )
{
	const unsigned long cr3  = pml4_table_create ( );
	const unsigned long pml4 = ( unsigned long ) VIRT ( cr3 );
	const unsigned long vm_pmem_pfn = PFN_DOWN_2MB ( PHYS ( vm_pmem_start ) );
	int i;

	for ( i = 0; i < PFN_UP_2MB ( vm_pmem_size ); i++ ) {
		const unsigned long vm_paddr = i << PAGE_SHIFT_2MB;
		const unsigned long pm_pfn   = i + ( is_reserved_pmem ( i ) ? 0 : vm_pmem_pfn ); /* [NAIVE] for VGA */
		const unsigned long pm_paddr = pm_pfn << PAGE_SHIFT_2MB;

		mmap ( pml4, vm_paddr, pm_paddr, 1 /* is_user */ );
	}

	printf ( "Shadow table created (vaddr=%x, size=%x).\n", vm_pmem_start, vm_pmem_size );
	return cr3;
}

static unsigned long 
create_initial_pg_table ( void )
{
	const unsigned long pages = 1 << 18; /* (1 << 18) pages = (1 << 30) bytes = 1 Gbytes */
	const unsigned long cr3   = pml4_table_create ( );
	const unsigned long pml4  = ( unsigned long ) VIRT ( cr3 );

	int i;
	for ( i = 0; i < pages; i++ ) {
		mmap ( pml4, i << PAGE_SHIFT, i << PAGE_SHIFT, 1 /* is_user */ );
	}

	return cr3;
}

void 
print_something ( )
{
	printf ( "\n\n" "Hello World." "\n\n" );
}

void
vm_create ( struct vm *vm, unsigned long guest_image_start, unsigned long guest_image_size, unsigned long vm_pmem_size )
{
	struct vmcb *vmcb;

	vmcb = alloc_vmcb ( );
	vm->vmcb = vmcb;

	vmcb->guest_asid  = 1;
        vmcb->tlb_control = 1;	
	vmcb->general2_intercepts = INTRCPT_VMRUN;
	vmcb->iopm_base_pa  = create_intercept_table ( 12 << 10 /* 12 Kbytes */ );
	vmcb->msrpm_base_pa = create_intercept_table ( 8 << 10  /* 8 Kbytes */ );

	printf ( "Nested paging is enabled -----.\n" );
	vmcb->np_enable = 1;
	{
		u32 eax, edx;
		rdmsr ( MSR_PAT, eax, edx );
		vmcb->g_pat = ( ( ( (u64) edx ) << 32 ) | eax );
	}


	/* [Note] vm_pmem_start is a virtual base address of the physical memory of the virtual machine.  */
	const unsigned long vm_pmem_start = alloc_vm_pmem ( vm_pmem_size );
	vmcb->h_cr3 = create_vm_pmem_mapping_table ( vm_pmem_start, vm_pmem_size );
	

	/* copy guest image */
	memmove ( ( void * ) vm_pmem_start, VIRT ( 0 ), vm_pmem_size );  

	vm->vmcb->rip = ( unsigned long )( &print_something );

	{
		unsigned long rsp;
		__asm__ __volatile__ ( "movq %%rsp, %0" : "=r" ( rsp ) );
		vm->vmcb->rsp = rsp;
	}

	{
		vmcb->cs.sel   = VMM_CS64;
//		vmcb->cs.attrs.bytes = 0x089b; // 1000 1001 1011
		vmcb->cs.attrs.fields.type = 0xa;
		vmcb->cs.attrs.fields.l = 1;
		vmcb->cs.attrs.fields.g = 1;
		vmcb->cs.limit = 0xffff;
		vmcb->cs.base  = 0;
	}

	// type = 1011 : Code-segment, R(eadable), A(ccessed)
	// S(ystem/user), P(resent), G(ranularity)

	int i;
	struct seg_selector *segregs [] = { &vmcb->ss, &vmcb->ds, &vmcb->es, &vmcb->fs, &vmcb->gs, NULL };
	for ( i = 0; segregs [ i ] != NULL; i++ ) {
		struct seg_selector *x = segregs [ i ];
		x->sel   = VMM_DS64;
		x->attrs.bytes = 0x0893; // 1000 1001 0011 
		x->limit = 0xffff;
		x->base  = 0;

		// type = 0011 : Data-segment, W(ritable), A(ccessed)
		// S(ystem/user), P(resent), G(ranularity)
	}

	{
		u64 dtr;
		__asm__ __volatile__ ( "sgdt %0":"=m" ( dtr ) );
		vmcb->gdtr.sel   = 0;
		vmcb->gdtr.attrs.bytes = 0;
		vmcb->gdtr.limit = dtr & 0xff;
		vmcb->gdtr.base  = dtr >> 16;
	}

	{
		unsigned long rflags;
		__asm__ __volatile__( "pushfq; popq %0" :"=m" ( rflags ) );
		vmcb->rflags = rflags & 2;
	}

	vmcb->cr0 = read_cr0 ();
	vmcb->cr3 = read_cr3 (); /* guest */
	vmcb->cr4 = read_cr4 ();

	{
		u32 eax, edx;
		rdmsr ( MSR_EFER, eax, edx );
		vmcb->efer = ( ( ( (u64) edx) << 32 ) | eax );
	}

	printf ( "New virtual machine created.\n" ); 
}

/******************************************************/

static void
switch_to_guest_os ( struct vm *vm )
{
	u64 p_vmcb = PHYS ( vm->vmcb );
	svm_launch ( p_vmcb );
}

static void
handle_vmexit ( struct vm *vm )
{
	print_vmexit_exitcode ( vm->vmcb->exitcode );
	printf ( "VMCB: rip=%x\n", vm->vmcb->rip );

	printf ( "fault address=%x, error_code=%x\n", vm->vmcb->exitinfo2, vm->vmcb->exitinfo1 );

	/* p. 268, p.490 */
	if ( vm->vmcb->exitinfo1 & 1 ) {
		printf ( "page fault was caused by a page-protection violation\n" );
	} else {
		printf ( "page fault was caused by a not-present page\n" );
	}

	if ( vm->vmcb->exitinfo1 & 2 ) {
		printf ( "memory access was write\n" );
	} else {
		printf ( "memory access was read\n" );
	}

	if ( vm->vmcb->exitinfo1 & 4 ) {
		printf ( "an access in supervisor mode caused the page fault\n" );
	} else {
		printf ( "an access in user mode caused the page fault\n" );
	}
}

void
vm_boot ( struct vm *vm )
{
	printf ( "Booting guest operating system...\n\n" ); 

	vmcb_check_consistency ( vm->vmcb );

	while ( 1 ) {
		switch_to_guest_os ( vm );

		/* #VMEXIT occures. */

		handle_vmexit ( vm );
		
		break; /* [DEBUG] */
	}
}
